import threading
import sys
import redis
import random
import string
import time

from datetime import datetime

from threading import Event
from libraries.security.user import AuraUser
from libraries.database.statestore import RedisStateStore
from modules.communication.redis.messenger import RedisMessenger
from libraries.exceptions.auraexceptions import RedisConnectionException
from libraries.enum.consolecolor import TerminalColors
from libraries.enum.redischannels import Channels


class ServerRedisAdapter(threading.Thread, RedisMessenger):
    debug = False
    pubsub = None
    config = None
    redisdb = None
    channel = ""
    auracontroller = None
    redisclient = None

    def __init__(self, debug=False):
        threading.Thread.__init__(self)
        RedisMessenger.__init__(self)

        self.debug = debug

        # init
        threading.Thread.__init__ (self)
        self.shutdown_event = Event()

        self.channel = Channels.STANDARD.value
        self.section = ''
        self.rstore = RedisStateStore()
        self.errnr = '00'
        self.components = {'controller':'01', 'scheduling':'02', 'playd':'03', 'recorder':'04', 'helpers':'09'}
        self.fromMail = ''
        self.adminMails = ''

        self.redisclient = ClientRedisAdapter()

    # ------------------------------------------------------------------------------------------ #
    def set_controller(self, auracontroller):
        self.auracontroller = auracontroller
        self.scheduler = auracontroller.scheduler

    # ------------------------------------------------------------------------------------------ #
    def set_config(self, config):
        self.config = config

    # ------------------------------------------------------------------------------------------ #
    def run(self):
        self.redisdb = redis.Redis()
        self.pubsub = self.redisdb.pubsub()
        self.pubsub.subscribe(self.channel)

        print(TerminalColors.OK_ORANGE.value + "waiting for REDIS message on channel " + self.channel + TerminalColors.ENDC.value)

        for item in self.pubsub.listen():
            if item["type"] == "subscribe":
                continue

            print(TerminalColors.OK_ORANGE.value + "received REDIS message: " + TerminalColors.ENDC.value + str(item) )

            item["channel"] = item["channel"].decode("utf-8")
            if isinstance(item["data"], bytes):
                item["data"] = item["data"].decode("utf-8")

            if item['data'] == "KILL":
                break

            try:
                self.work(item)
            except RedisConnectionException as rce:
                print(rce)
                #traceback.print_exc()

            print(TerminalColors.OK_ORANGE.value + "waiting for REDIS message on channel " + self.channel + TerminalColors.ENDC.value)

        self.pubsub.unsubscribe()
        print(self, "unsubscribed from " + self.channel + " and finished")

    # ------------------------------------------------------------------------------------------ #
    def listen_for_one_message(self, channel, socket_timeout=2):
        self.redisdb = redis.Redis(socket_timeout=socket_timeout)
        self.pubsub = self.redisdb.pubsub()
        self.pubsub.subscribe(channel)

        try:
            if self.debug:
                print("I am listening on channel '"+channel+"' for "+str(socket_timeout)+" seconds")

            for item in self.pubsub.listen():
                it = self.receive_message(item)
                if it is not None:
                    break
        except redis.exceptions.TimeoutError as te:
            raise te

        return item["data"]

    # ------------------------------------------------------------------------------------------ #
    def receive_message(self, item):
        if item["type"] == "subscribe":
            if self.debug > 2:
                print("i am subscribed to channel " + item["channel"].decode("utf-8"))
            return None

        item["channel"] = item["channel"].decode("utf-8")
        if isinstance(item["data"], bytes):
            item["data"] = item["data"].decode("utf-8")

        self.pubsub.unsubscribe()
        return item

    # ------------------------------------------------------------------------------------------ #
    def work(self, item):
        if item["data"] == "fetch_new_programme":
            self.execute(self.auracontroller.fetch_new_programme, Channels.FNP_REPLY.value)

        elif item["data"] == "init_player":
            self.execute(self.auracontroller.init_player, Channels.IP_REPLY.value)

        elif item["data"] == "get_act_programme":
            self.execute(self.auracontroller.get_act_programme, Channels.GAP_REPLY.value)

        elif item["data"] == "print_message_queue":
            self.execute(self.scheduler.print_message_queue, Channels.PMQ_REPLY.value)

        else:
            raise RedisConnectionException("ServerRedisAdapter Cannot understand command: " + item["data"])

    # ------------------------------------------------------------------------------------------ #
    def execute(self, f, channel):
        reply = f()

        if reply is None:
            reply = ""

        # sometimes the sender is faster than the receiver and with redis messages would be lost
        time.sleep(0.1)

        # publish
        self.redisclient.publish(channel, reply)

    # ------------------------------------------------------------------------------------------ #
    def join_comm(self):
        try:
            while self.is_alive():
                print(str(datetime.now())+" joining")
                self.join()
                print("join out")

                # if cnt % 30 == 0:
                #    print(datetime.datetime.now().isoformat())
                #    server.printLastMessages()
                #    cnt = 0
                # cnt = cnt + 1

        except (KeyboardInterrupt, SystemExit):
            # Dem Server den Shutdown event setzen
            # server.shutdown_event.set()
            # Der Server wartet auf Eingabe
            # Daher einen Client initiieren, der eine Nachricht schickt
            self.halt()
            sys.exit('Terminated')

    # ------------------------------------------------------------------------------------------ #
    def halt(self):
        """
        Stop the server
        """
        if self.shutdown_event.is_set():
            return
        try:
            del self.auracontroller
        except:
            pass
        self.shutdown_event.set()
        result = 'failed'
        try:
            result = self.socket.unbind("tcp://"+self.ip+":"+self.port)
        except:
            pass

        #self.socket.close()

    # ------------------------------------------------------------------------------------------ #
    def send(self, message):
        """
        Send a message to the client
        :param message: string
        """
        if not self.can_send:
            print("sending a "+str(len(message))+" long message via ZMQ.")
            self.socket.send(message.encode("utf-8"))
            self.can_send = False
        else:
            print("cannot send message via ZMQ: "+str(message))

    # ------------------------------------------------------------------------------------------ #
    @staticmethod
    def get_accounts(self):
        """
        Get accounts from redis db
        :return: llist - a list of accounts
        """
        accounts = AuraUser().getLogins()
        db = redis.Redis()

        internaccount = db.get('internAccess')
        if not internaccount:
            user = ''.join(random.sample(string.ascii_lowercase,10))
            password = ''.join(random.sample(string.ascii_lowercase+string.ascii_uppercase+string.digits,22))
            db.set('internAccess', user + ':' + password)
            intern = [user, password]
        else:
            intern =  internaccount.split(':')

        accounts[intern[0]] = intern[1]

        return accounts


class ClientRedisAdapter(RedisMessenger):
    debug = False

    def __init__(self, debug=True):
        RedisMessenger.__init__(self)
        self.debug = debug

    # ------------------------------------------------------------------------------------------ #
    def publish(self, channel, message):
        if type(channel) == Channels:
            channel = channel.value

        self.rstore.publish(channel, message)
