import threading
import zmq
import time
import sys
import zmq.auth
import redis
import random
import string
import simplejson
import traceback

from datetime import datetime

from zmq.auth.thread import ThreadAuthenticator
from threading import Event
from libraries.security.whitelist import AuraWhitelist
from libraries.security.user import AuraUser
from modules.scheduling.scheduler import AuraScheduler


class ServerZMQAdapter(threading.Thread):
    debug = False
    config = None
    auracontroller = None

    def __init__(self, controller, config, debug):
        self.auracontroller = controller
        self.ip = config.get('zmqhostip')
        self.port = config.get('zmqport')
        self.debug = debug

        # init
        threading.Thread.__init__ (self)
        self.shutdown_event = Event()

        self.config = config

        self.context = zmq.Context().instance()
        self.authserver = ThreadAuthenticator(self.context)
        self.socket = self.context.socket(zmq.REP)
        self.can_send = False

        # start thread
        self.start()

    # ------------------------------------------------------------------------------------------ #
    def run(self):
        """
        run runs on function start
        """
        self.auracontroller.set_sender(self)
        self.start_auth_server()
        self.socket.plain_server = True
#        print("binding to: tcp://"+self.ip+":"+self.port)
        self.socket.bind("tcp://"+self.ip+":"+self.port)
        self.shutdown_event.clear()
        self.auracontroller.messenger.set_mail_addresses(self.config.get('frommail'), self.config.get('adminmail'))

        # Process tasks forever
        while not self.shutdown_event.is_set():
            print("waiting for zmq message on socket tcp://"+self.ip+":"+self.port+"\n")
            data = self.socket.recv().decode("utf-8")

            msg = ''
            retval = None

            self.can_send = True

            datasplit = data.split(' ')

            command = str(datasplit.pop(0))
            params = "()" if len(datasplit) < 1 else  "'" + "','".join(datasplit) + "'"

            try:
                statement = "self.auracontroller." + command + params

                if self.debug:
                    print("zmq received message '" + data + "' and trying to call " + statement)

                if params == "()":
                    # if there are no params => easily use getattr and get a retval
                    f = getattr(self.auracontroller, command)
                    retval = f()  # json string
                else:
                    # if there are params => use exec and get no retval
                    exec("retval = " + statement)

            except SyntaxError as e:
                traceback.print_exc()

                msg = 'Warning: Syntax Error'
                self.auracontroller.message(msg)
            except AttributeError as e:
                traceback.print_exc()

                msg = 'Warning: Method ' + command + ' does not exist'
                self.auracontroller.message(msg)
            except TypeError as e:
                traceback.print_exc()

                msg = 'Warning: Wrong number of params'
                self.auracontroller.message(msg)
            except Exception as e:
                traceback.print_exc()
                msg = 'Warning: Unknown Error'
                self.auracontroller.message(msg)

            if msg == '':
                msg = 'OK'

            if retval is None:
                self.socket.send(msg.encode())
            else:
                self.socket.send(simplejson.dumps(retval).encode())

        return

    # ------------------------------------------------------------------------------------------ #
    def join_comm(self):
        try:
            while self.is_alive():
                print(str(datetime.now())+" joining")
                self.join()
                print("join out")

                # if cnt % 30 == 0:
                #    print(datetime.datetime.now().isoformat())
                #    server.printLastMessages()
                #    cnt = 0
                # cnt = cnt + 1

        except (KeyboardInterrupt, SystemExit):
            # Dem Server den Shutdown event setzen
            # server.shutdown_event.set()
            # Der Server wartet auf Eingabe
            # Daher einen Client initiieren, der eine Nachricht schickt
            self.halt()
            sys.exit('Terminated')

    # ------------------------------------------------------------------------------------------ #
    def halt(self):
        """
        Stop the server
        """
        if self.shutdown_event.is_set():
            return
        try:
            del self.auracontroller
        except:
            pass
        self.shutdown_event.set()
        result = 'failed'
        try:
            result = self.socket.unbind("tcp://"+self.ip+":"+self.port)
        except:
            pass
        #self.socket.close()

    # ------------------------------------------------------------------------------------------ #
    def reload(self):
        """
        stop, reload config and startagaing
        """
        if self.shutdown_event.is_set():
            return
        self.loadConfig()
        self.halt()
        time.sleep(3)
        self.run()

    # ------------------------------------------------------------------------------------------ #
    def send(self, message):
        """
        Send a message to the client
        :param message: string
        """
        if not self.can_send:
            print("sending a "+str(len(message))+" long message via ZMQ.")
            self.socket.send(message.encode("utf-8"))
            self.can_send = False
        else:
            print("cannot send message via ZMQ: "+str(message))

    # ------------------------------------------------------------------------------------------ #
    def start_auth_server(self):
        """
        Start zmq authentification server
        """
        # stop auth server if running
        if self.authserver.is_alive():
            self.authserver.stop()

        if self.config.get("securitylevel") > 0:
            # Authentifizierungsserver starten.

            self.authserver.start()

            # Bei security level 2 auch passwort und usernamen verlangen
            if self.securitylevel > 1:
                try:

                    addresses = AuraWhitelist().getList()
                    for address in addresses:
                        self.authserver.allow(address)

                except:
                    print("Exception passed!!!")
                    pass

            # Instruct authenticator to handle PLAIN requests
            self.authserver.configure_plain(domain='*', passwords=self.get_accounts())

    # ------------------------------------------------------------------------------------------ #
    @staticmethod
    def get_accounts(self):
        """
        Get accounts from redis db
        :return: llist - a list of accounts
        """
        accounts = AuraUser().getLogins()
        db = redis.Redis()

        internaccount = db.get('internAccess')
        if not internaccount:
            user = ''.join(random.sample(string.ascii_lowercase,10))
            password = ''.join(random.sample(string.ascii_lowercase+string.ascii_uppercase+string.digits,22))
            db.set('internAccess', user + ':' + password)
            intern = [user, password]
        else:
            intern =  internaccount.split(':')

        accounts[intern[0]] = intern[1]

        return accounts


class ClientZMQAdapter:
    debug = False

    def __init__(self, host, port, debug=True):
        self.host = host
        self.port = port
        self.debug = debug

        self.username = ""
        self.password = ""

        # Socket to receive messages on
        context = zmq.Context()
        self.socket = context.socket(zmq.REQ)

        self.socket.RCVTIMEO = 10000
        self.message = ''

    # ------------------------------------------------------------------------------------------ #
    def send(self, message, blocking=True):
        """
        Send a message to the server
        :param message: string
        :return:boolean
        """
        db = redis.Redis()
        account = db.get('internAccess')
        (username, password) = account.decode().split(':')

        self.socket.plain_username = username.encode()
        self.socket.plain_password = password.encode()
        self.socket.connect("tcp://"+self.host+":"+self.port)
        self.socket.send_string(str(message))

        try:
            if self.debug:
                if blocking:
                    print("message '" + str(message) + "' sent to auraserver and waiting for reply")
                else:
                    print("message '" + str(message) + "' sent to auraserver. i am not waiting for a reply")

            if blocking:
                self.message = self.socket.recv().decode("utf-8")
            else:
                self.socket.recv(zmq.NOBLOCK)

            if self.debug:
                print("got reply from '"+str(message)+"': "+self.message)

        except zmq.Again as e:
            self.message = 'Not waiting for a reply'
        except zmq.ZMQError as e:
            self.message = ''
            print(e)
            return False
        except Exception as e:
            self.message = ''
            print(e)
            return False

        return self.message

    # ------------------------------------------------------------------------------------------ #
    def set_username(self, username):
        """
        authentification to the server with username
        :param username: string - the username
        """
        self.username = username

    # ------------------------------------------------------------------------------------------ #
    def set_password(self, password):
        """
        authentification to the server with password
        :param password: string - the password
        """
        self.password = password

    # ------------------------------------------------------------------------------------------ #
    def receive(self):
        """
        Return the currently received message
        :return:string - message from server
        """
        return self.message

